//
// Copyright (c) 2003-2009, MindFusion LLC - Bulgaria.
//

using System;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;

using MindFusion.Diagramming;
using MindFusion.Diagramming.Layout;
using MindFusion.Diagramming.WinForms;

using Orientation = MindFusion.Diagramming.Orientation;
using SolidBrush = MindFusion.Drawing.SolidBrush;


namespace MindFusion.Diagramming.WinForms.Samples.CS.WorkflowDesigner
{
	/// <summary>
	/// Summary description for Form1.
	/// </summary>
	public class MainForm : Form
	{
		#region Form

		private DiagramView view;
		private Diagram diagram;
		private ListBox palette;
		private Button btnNew;
		private Button btnLayout;
		private Button btnLoad;
		private Button btnSave;
		private Button btnUndo;
		private Button btnRedo;

		/// <summary>
		/// Required designer variable.
		/// </summary>
		private Container components = null;

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main()
		{
			Application.Run(new MainForm());
		}

		public MainForm()
		{
			//
			// Required for Windows Form Designer support
			//
			InitializeComponent();

			//
			// TODO: Add any constructor code after InitializeComponent call
			//
		}

		private void MainForm_Load(object sender, EventArgs e)
		{
			view.InplaceEditAcceptOnEnter = true;
			view.DelKeyAction = DelKeyAction.None;

			diagram.LinkHeadShape = ArrowHeads.Triangle;
			diagram.LinkHeadShapeSize = 3;
			diagram.LinkBrush = new SolidBrush(Color.LightGray);

			diagram.UndoManager.UndoEnabled = true;
			diagram.RouteLinks = true;

			NewWorkflow();
		}

		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				if (components != null)
				{
					components.Dispose();
				}
			}
			base.Dispose(disposing);
		}

		private void btnNew_Click(object sender, EventArgs e)
		{
			NewWorkflow();
		}

		private void btnLayout_Click(object sender, EventArgs e)
		{
			LayoutDiagram();
		}

		private void btnSave_Click(object sender, EventArgs e)
		{
			SaveFileDialog sfd = new SaveFileDialog();
			if (sfd.ShowDialog() == DialogResult.OK)
				diagram.SaveToFile(sfd.FileName);
		}

		private void btnLoad_Click(object sender, EventArgs e)
		{
			OpenFileDialog ofd = new OpenFileDialog();
			if (ofd.ShowDialog() == DialogResult.OK)
				diagram.LoadFromFile(ofd.FileName);
		}

		#endregion

		#region Windows Form Designer generated code

		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			System.ComponentModel.ComponentResourceManager resources = new System.ComponentModel.ComponentResourceManager(typeof(MainForm));
			this.view = new MindFusion.Diagramming.WinForms.DiagramView();
			this.diagram = new MindFusion.Diagramming.Diagram();
			this.palette = new System.Windows.Forms.ListBox();
			this.btnNew = new System.Windows.Forms.Button();
			this.btnLayout = new System.Windows.Forms.Button();
			this.btnLoad = new System.Windows.Forms.Button();
			this.btnSave = new System.Windows.Forms.Button();
			this.btnUndo = new System.Windows.Forms.Button();
			this.btnRedo = new System.Windows.Forms.Button();
			this.SuspendLayout();
			// 
			// view
			// 
			this.view.AllowDrop = true;
			this.view.Anchor = ((System.Windows.Forms.AnchorStyles)((((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Bottom)
						| System.Windows.Forms.AnchorStyles.Left)
						| System.Windows.Forms.AnchorStyles.Right)));
			this.view.Behavior = MindFusion.Diagramming.Behavior.LinkShapes;
			this.view.ControlHandlesStyle = MindFusion.Diagramming.HandlesStyle.HatchHandles;
			this.view.ControlMouseAction = MindFusion.Diagramming.ControlMouseAction.SelectNode;
			this.view.DelKeyAction = MindFusion.Diagramming.DelKeyAction.DeleteSelectedItems;
			this.view.Diagram = this.diagram;
			this.view.Location = new System.Drawing.Point(12, 12);
			this.view.MiddleButtonActions = MindFusion.Diagramming.MouseButtonActions.None;
			this.view.ModificationStart = MindFusion.Diagramming.ModificationStart.SelectedOnly;
			this.view.Name = "view";
			this.view.RightButtonActions = MindFusion.Diagramming.MouseButtonActions.Cancel;
			this.view.Size = new System.Drawing.Size(726, 729);
			this.view.TabIndex = 0;
			this.view.Text = "diagramView1";
			this.view.DragOver += new System.Windows.Forms.DragEventHandler(this.view_DragOver);
			this.view.DragDrop += new System.Windows.Forms.DragEventHandler(this.view_DragDrop);
			// 
			// diagram
			// 
			this.diagram.BackBrush = new MindFusion.Drawing.SolidBrush("#FFFFFFFF");
			this.diagram.LinkBaseShape = null;
			this.diagram.LinkIntermediateShape = null;
			this.diagram.LinkRouter = new MindFusion.Diagramming.QuickRouter();
			// 
			// palette
			// 
			this.palette.Anchor = ((System.Windows.Forms.AnchorStyles)(((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Bottom)
						| System.Windows.Forms.AnchorStyles.Right)));
			this.palette.Items.AddRange(new object[] {
            "Activity",
            "IfElse",
            "While",
            "Process"});
			this.palette.Location = new System.Drawing.Point(744, 12);
			this.palette.Name = "palette";
			this.palette.Size = new System.Drawing.Size(132, 498);
			this.palette.TabIndex = 1;
			this.palette.MouseMove += new System.Windows.Forms.MouseEventHandler(this.palette_MouseMove);
			this.palette.MouseDown += new System.Windows.Forms.MouseEventHandler(this.palette_MouseDown);
			// 
			// btnNew
			// 
			this.btnNew.Anchor = ((System.Windows.Forms.AnchorStyles)((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Right)));
			this.btnNew.Location = new System.Drawing.Point(776, 589);
			this.btnNew.Name = "btnNew";
			this.btnNew.Size = new System.Drawing.Size(75, 23);
			this.btnNew.TabIndex = 2;
			this.btnNew.Text = "New";
			this.btnNew.Click += new System.EventHandler(this.btnNew_Click);
			// 
			// btnLayout
			// 
			this.btnLayout.Anchor = ((System.Windows.Forms.AnchorStyles)((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Right)));
			this.btnLayout.Location = new System.Drawing.Point(776, 621);
			this.btnLayout.Name = "btnLayout";
			this.btnLayout.Size = new System.Drawing.Size(75, 23);
			this.btnLayout.TabIndex = 3;
			this.btnLayout.Text = "Layout";
			this.btnLayout.Click += new System.EventHandler(this.btnLayout_Click);
			// 
			// btnLoad
			// 
			this.btnLoad.Anchor = ((System.Windows.Forms.AnchorStyles)((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Right)));
			this.btnLoad.Location = new System.Drawing.Point(776, 709);
			this.btnLoad.Name = "btnLoad";
			this.btnLoad.Size = new System.Drawing.Size(75, 23);
			this.btnLoad.TabIndex = 5;
			this.btnLoad.Text = "Load";
			this.btnLoad.Click += new System.EventHandler(this.btnLoad_Click);
			// 
			// btnSave
			// 
			this.btnSave.Anchor = ((System.Windows.Forms.AnchorStyles)((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Right)));
			this.btnSave.Location = new System.Drawing.Point(776, 677);
			this.btnSave.Name = "btnSave";
			this.btnSave.Size = new System.Drawing.Size(75, 23);
			this.btnSave.TabIndex = 4;
			this.btnSave.Text = "Save";
			this.btnSave.Click += new System.EventHandler(this.btnSave_Click);
			// 
			// btnUndo
			// 
			this.btnUndo.Location = new System.Drawing.Point(744, 530);
			this.btnUndo.Name = "btnUndo";
			this.btnUndo.Size = new System.Drawing.Size(75, 23);
			this.btnUndo.TabIndex = 6;
			this.btnUndo.Text = "Undo";
			this.btnUndo.UseVisualStyleBackColor = true;
			this.btnUndo.Click += new System.EventHandler(this.btnUndo_Click);
			// 
			// btnRedo
			// 
			this.btnRedo.Location = new System.Drawing.Point(801, 548);
			this.btnRedo.Name = "btnRedo";
			this.btnRedo.Size = new System.Drawing.Size(75, 23);
			this.btnRedo.TabIndex = 7;
			this.btnRedo.Text = "Redo";
			this.btnRedo.UseVisualStyleBackColor = true;
			this.btnRedo.Click += new System.EventHandler(this.btnRedo_Click);
			// 
			// MainForm
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(888, 753);
			this.Controls.Add(this.btnRedo);
			this.Controls.Add(this.btnUndo);
			this.Controls.Add(this.btnLoad);
			this.Controls.Add(this.btnSave);
			this.Controls.Add(this.btnLayout);
			this.Controls.Add(this.btnNew);
			this.Controls.Add(this.palette);
			this.Controls.Add(this.view);
			this.Icon = ((System.Drawing.Icon)(resources.GetObject("$this.Icon")));
			this.Name = "MainForm";
			this.StartPosition = System.Windows.Forms.FormStartPosition.CenterScreen;
			this.Text = "MindFusion.Diagramming Sample: Workflow Designer";
			this.Load += new System.EventHandler(this.MainForm_Load);
			this.ResumeLayout(false);

		}

		#endregion


		#region Workflow creation

		private void NewWorkflow()
		{
			diagram.Tag = 0;
			diagram.ClearAll();

			ShapeNode start = diagram.Factory.CreateShapeNode(defaultPosition, Shapes.Start);
			start.Text = "Start";
			start.Brush = new SolidBrush(Color.LightGreen);
			start.LayoutTraits[WorkflowItemId] = "Start";
			start.LayoutTraits[PairStart] = diagram.Tag;

			ShapeNode stop = diagram.Factory.CreateShapeNode(defaultPosition, Shapes.Terminator);
			stop.Text = "Stop";
			stop.Brush = new SolidBrush(Color.Red);
			stop.LayoutTraits[PairEnd] = diagram.Tag;

			CreateLink(start, stop, true, "ControlFlow");

			view.Behavior = Behavior.Modify;

			LayoutDiagram();
			diagram.Tag = (int)diagram.Tag + 1;
		}

		private void CreateActivity(DiagramLink link)
		{
			ShapeNode origin = link.Origin as ShapeNode;
			ShapeNode dest = link.Destination as ShapeNode;

			string oldLogicID = link.LayoutTraits[Connector].ToString();
			diagram.Links.Remove(link);

			ShapeNode newActivity = diagram.Factory.CreateShapeNode(defaultPosition, Shapes.Process);
			newActivity.LayoutTraits[WorkflowItemId] = "Activity";

			CreateLink(origin, newActivity, true, oldLogicID);
			CreateLink(newActivity, dest, true, "ControlFlow");

			LayoutDiagram();

			view.BeginEdit(newActivity);
		}

		private void CreateProcess(DiagramLink link)
		{
			CreateActivity(link);
			DiagramLink lastLink = diagram.Links[diagram.Links.Count - 1];
			CreateIfElse(lastLink);
			DiagramLink trueLink = diagram.Links[diagram.Links.Count - 3];
			DiagramLink falseLink = diagram.Links[diagram.Links.Count - 2];
			CreateActivity(trueLink);
			CreateActivity(falseLink);
		}

		private void CreateIfElse(DiagramLink link)
		{
			ShapeNode origin = link.Origin as ShapeNode;
			ShapeNode dest = link.Destination as ShapeNode;

			string oldLogicID = link.LayoutTraits[Connector].ToString();
			diagram.Links.Remove(link);

			ShapeNode decision = diagram.Factory.CreateShapeNode(defaultPosition, Shapes.Decision);
			decision.Text = "Condition";
			decision.LayoutTraits[WorkflowItemId] = "IfElse";
			decision.LayoutTraits[PairStart] = diagram.Tag;

			ShapeNode endIf = diagram.Factory.CreateShapeNode(smallSize, Shapes.Ellipse);
			endIf.Text = "";
			endIf.LayoutTraits[PairEnd] = diagram.Tag;

			CreateLink(origin, decision, true, oldLogicID);
			CreateLink(decision, endIf, true, "Yes");
			CreateLink(decision, endIf, true, "No");
			CreateLink(endIf, dest, true, "ControlFlow");

			LayoutDiagram();
			diagram.Tag = (int)diagram.Tag + 1;
		}

		private void CreateWhile(DiagramLink link)
		{
			ShapeNode origin = link.Origin as ShapeNode;
			ShapeNode dest = link.Destination as ShapeNode;

			string oldLogicID = link.LayoutTraits[Connector].ToString();
			diagram.Links.Remove(link);

			ShapeNode loopStart = diagram.Factory.CreateShapeNode(defaultPosition, Shapes.BeginLoop);
			loopStart.LayoutTraits[WorkflowItemId] = "While";
			loopStart.Text = "while";
			loopStart.LayoutTraits[PairStart] = diagram.Tag;

			ShapeNode loopEnd = diagram.Factory.CreateShapeNode(defaultPosition, Shapes.EndLoop);
			loopEnd.Text = "end while";
			loopEnd.LayoutTraits[PairEnd] = diagram.Tag;

			CreateLink(origin, loopStart, true, oldLogicID);
			CreateLink(loopStart, loopEnd, true, "ControlFlow");
			CreateLink(loopEnd, dest, true, "ControlFlow");

			DiagramLink backLink = CreateLink(loopEnd, loopStart, false, "While");
			backLink.IgnoreLayout = true;

			LayoutDiagram();
			diagram.Tag = (int)diagram.Tag + 1;
		}

		private DiagramLink CreateLink(ShapeNode n1, ShapeNode n2,
			bool acceptActivities, string logicID)
		{
			DiagramLink link = diagram.Factory.CreateDiagramLink(n1, n2);
			link.Tag = acceptActivities;
			if (acceptActivities)
				link.Text = "drop here";

			link.LayoutTraits[Connector] = logicID;

			return link;
		}

		#endregion

		#region Workflow layout

		private void LayoutDiagram()
		{
			view.EndEdit(true);

			LayoutBranch(diagram.Nodes[0], 0);
			diagram.ResizeToFitItems(10);
		}

		private void LayoutBranch(DiagramNode node, int branchEnd)
		{
			if (node.LayoutTraits[PairEnd] != null &&
				(int)node.LayoutTraits[PairEnd] == branchEnd)
				return;

			DiagramNode continueFrom = null;

			int endMarker = node.LayoutTraits[PairStart] == null ?
				branchEnd : (int)node.LayoutTraits[PairStart];

			if ("IfElse".Equals(node.LayoutTraits[WorkflowItemId]))
			{
				DiagramLink trueLink = GetOutLink(node, "Yes");
				DiagramLink falseLink = GetOutLink(node, "No");
				continueFrom = GetPairEnd(endMarker);

				// layout the "true" branch at the left side under the decision node
				if (trueLink.Destination != continueFrom)
				{
					float rightWidth = GetRightWidth(trueLink.Destination, endMarker);
					DiagramNode branchStart = trueLink.Destination;
					branchStart.Bounds = new RectangleF(node.Bounds.Left +
						node.Bounds.Width / 2 - rightWidth - branchStart.Bounds.Width / 2 - 5,
						node.Bounds.Bottom + LevelDistance, branchStart.Bounds.Width, branchStart.Bounds.Height);
					LayoutBranch(branchStart, endMarker);
				}

				// layout the "false" branch at the right side under the decision node
				if (falseLink.Destination != continueFrom)
				{
					float leftWidth = GetLeftWidth(falseLink.Destination, endMarker);
					DiagramNode branchStart = falseLink.Destination;
					branchStart.Bounds = new RectangleF(node.Bounds.Right -
						node.Bounds.Width / 2 + leftWidth - branchStart.Bounds.Width / 2 + 5,
						node.Bounds.Bottom + LevelDistance, branchStart.Bounds.Width, branchStart.Bounds.Height);
					LayoutBranch(branchStart, endMarker);
				}

				// continue the layout below the longer branch
				RectangleF bb = continueFrom.Bounds;
				DiagramNode e1 = continueFrom.IncomingLinks[0].Origin;
				DiagramNode e2 = continueFrom.IncomingLinks[1].Origin;
				continueFrom.Bounds = new RectangleF(
					node.Bounds.Left + node.Bounds.Width / 2 - bb.Width / 2,
					Math.Max(e1.Bounds.Bottom, e2.Bounds.Bottom) + LevelDistance,
					bb.Width, bb.Height);

				// bend the "true" link
				if (trueLink.Destination != continueFrom)
					BendLinkLeftDown(trueLink);
				else
					BendLinkLeftDownRight(trueLink);

				// bend the "false" link
				if (falseLink.Destination != continueFrom)
					BendLinkRightDown(falseLink);
				else
					BendLinkRightDownLeft(falseLink);

				// bend the branch ending links
				for (int l = 0; l < 2; ++l)
				{
					if (continueFrom.IncomingLinks[l].Origin != node)
					{
						DiagramLink link = continueFrom.IncomingLinks[l];
						BendLinkDownLOR(link);
					}
				}
			}
			else if ("Activity".Equals(node.LayoutTraits[WorkflowItemId]) ||
				"Start".Equals(node.LayoutTraits[WorkflowItemId]) ||
				node.LayoutTraits[WorkflowItemId] == null)
			{
				continueFrom = GetOutLink(node, "ControlFlow").Destination;
				RectangleF bb = continueFrom.Bounds;
				continueFrom.Bounds = new RectangleF(
					node.Bounds.Left + node.Bounds.Width / 2 - bb.Width / 2,
					node.Bounds.Bottom + LevelDistance,
					bb.Width, bb.Height);

				DiagramLink link = GetOutLink(node, "ControlFlow");
				SetLinkStraight(link);
			}
			else if ("While".Equals(node.LayoutTraits[WorkflowItemId]))
			{
				DiagramNode branchStart = GetOutLink(node, "ControlFlow").Destination;
				RectangleF bb = branchStart.Bounds;
				branchStart.Bounds = new RectangleF(
					node.Bounds.Left + node.Bounds.Width / 2 - bb.Width / 2,
					node.Bounds.Bottom + LevelDistance,
					bb.Width, bb.Height);
				LayoutBranch(branchStart, endMarker);

				DiagramLink link = GetOutLink(node, "ControlFlow");
				SetLinkStraight(link);

				float rightWidth = GetRightWidth(branchStart, endMarker);
				continueFrom = GetPairEnd(endMarker);

				DiagramLink backLink = GetOutLink(continueFrom, "While");
				BendBackLink(backLink, rightWidth);
			}

			if (continueFrom != null)
				LayoutBranch(continueFrom, branchEnd);
		}

		private DiagramNode GetPairEnd(int pairId)
		{
			foreach (DiagramNode node in diagram.Nodes)
			{
				if (node.LayoutTraits[PairEnd] != null &&
					(int)node.LayoutTraits[PairEnd] == pairId)
					return node;
			}

			return null;
		}

		private DiagramLink GetOutLink(DiagramNode node, string id)
		{
			foreach (DiagramLink link in node.OutgoingLinks)
			{
				if (link.LayoutTraits[Connector] is string &&
					(string)link.LayoutTraits[Connector] == id)
					return link;
			}

			return null;
		}

		// should merge the next 3 methods into a single one,
		// and calculate the width according to a parameter

		private float GetBranchWidth(DiagramNode node, int branchEnd)
		{
			float branchWidth = node.Bounds.Width;
			if (node.LayoutTraits[PairEnd] != null &&
				(int)node.LayoutTraits[PairEnd] == branchEnd)
				return branchWidth;

			int endMarker = node.LayoutTraits[PairStart] == null ? branchEnd : (int)node.LayoutTraits[PairStart];

			if ("IfElse".Equals(node.LayoutTraits[WorkflowItemId]))
			{
				DiagramLink trueLink = GetOutLink(node, "Yes");
				DiagramLink falseLink = GetOutLink(node, "No");
				branchWidth = Math.Max(node.Bounds.Width, 10 +
					GetBranchWidth(trueLink.Destination, endMarker) +
						GetBranchWidth(falseLink.Destination, endMarker)) + 10;
			}
			else if ("Activity".Equals(node.LayoutTraits[WorkflowItemId]) ||
				node.LayoutTraits[WorkflowItemId] == null)
			{
				branchWidth = Math.Max(node.Bounds.Width,
					GetBranchWidth(GetOutLink(node, "ControlFlow").Destination, endMarker));
			}
			else if ("While".Equals(node.LayoutTraits[WorkflowItemId]))
			{
				branchWidth = 5 + Math.Max(node.Bounds.Width,
					GetBranchWidth(GetOutLink(node, "ControlFlow").Destination, endMarker));
			}

			if (node.LayoutTraits[PairStart] != null)
			{
				int branchEndIdx = (int)node.LayoutTraits[PairStart];
				if (branchEndIdx != branchEnd)
				{
					branchWidth = Math.Max(branchWidth,
						GetBranchWidth(GetPairEnd(branchEndIdx), branchEnd));
				}
			}

			return branchWidth;
		}

		private float GetLeftWidth(DiagramNode node, int branchEnd)
		{
			float branchWidth = node.Bounds.Width / 2;
			if (node.LayoutTraits[PairEnd] != null &&
				(int)node.LayoutTraits[PairEnd] == branchEnd)
				return branchWidth;

			int endMarker = node.LayoutTraits[PairStart] == null ? branchEnd : (int)node.LayoutTraits[PairStart];

			if ("IfElse".Equals(node.LayoutTraits[WorkflowItemId]))
			{
				DiagramLink trueLink = GetOutLink(node, "Yes");
				branchWidth = Math.Max(node.Bounds.Width / 2, 5 +
					GetBranchWidth(trueLink.Destination, endMarker)) + 5;
			}
			else if ("Activity".Equals(node.LayoutTraits[WorkflowItemId]) ||
				node.LayoutTraits[WorkflowItemId] == null)
			{
				branchWidth = Math.Max(node.Bounds.Width / 2,
					GetLeftWidth(GetOutLink(node, "ControlFlow").Destination, endMarker));
			}
			else if ("While".Equals(node.LayoutTraits[WorkflowItemId]))
			{
				branchWidth = Math.Max(node.Bounds.Width / 2,
					GetLeftWidth(GetOutLink(node, "ControlFlow").Destination, endMarker));
			}

			if (node.LayoutTraits[PairStart] != null)
			{
				int branchEndIdx = (int)node.LayoutTraits[PairStart];
				if (branchEndIdx != branchEnd)
				{
					branchWidth = Math.Max(branchWidth,
						GetLeftWidth(GetPairEnd(branchEndIdx), branchEnd));
				}
			}

			return branchWidth;
		}

		private float GetRightWidth(DiagramNode node, int branchEnd)
		{
			float branchWidth = node.Bounds.Width / 2;
			if (node.LayoutTraits[PairEnd] != null &&
				(int)node.LayoutTraits[PairEnd] == branchEnd)
				return branchWidth;

			int endMarker = node.LayoutTraits[PairStart] == null ? branchEnd : (int)node.LayoutTraits[PairStart];

			if ("IfElse".Equals(node.LayoutTraits[WorkflowItemId]))
			{
				DiagramLink falseLink = GetOutLink(node, "No");
				branchWidth = Math.Max(node.Bounds.Width / 2, 5 +
					GetBranchWidth(falseLink.Destination, endMarker)) + 5;
			}
			else if ("Activity".Equals(node.LayoutTraits[WorkflowItemId]) ||
				node.LayoutTraits[WorkflowItemId] == null)
			{
				branchWidth = Math.Max(node.Bounds.Width / 2,
					GetRightWidth(GetOutLink(node, "ControlFlow").Destination, endMarker));
			}
			else if ("While".Equals(node.LayoutTraits[WorkflowItemId]))
			{
				branchWidth = 5 + Math.Max(node.Bounds.Width / 2,
					GetRightWidth(GetOutLink(node, "ControlFlow").Destination, endMarker));
			}

			if (node.LayoutTraits[PairStart] != null)
			{
				int branchEndIdx = (int)node.LayoutTraits[PairStart];
				if (branchEndIdx != branchEnd)
				{
					branchWidth = Math.Max(branchWidth,
						GetRightWidth(GetPairEnd(branchEndIdx), branchEnd));
				}
			}

			return branchWidth;
		}

		private PointF GetLeftMiddle(DiagramNode node)
		{
			return new PointF(
				node.Bounds.Left, node.Bounds.Top + node.Bounds.Height / 2);
		}

		private PointF GetRightMiddle(DiagramNode node)
		{
			return new PointF(
				node.Bounds.Right, node.Bounds.Top + node.Bounds.Height / 2);
		}

		private PointF GetTopMiddle(DiagramNode node)
		{
			return new PointF(
				node.Bounds.Left + node.Bounds.Width / 2, node.Bounds.Top);
		}

		private PointF GetBottomMiddle(DiagramNode node)
		{
			return new PointF(
				node.Bounds.Left + node.Bounds.Width / 2, node.Bounds.Bottom);
		}

		private void BendLinkLeftDown(DiagramLink link)
		{
			link.CascadeOrientation = Orientation.Horizontal;
			link.Style = LinkStyle.Cascading;
			link.SegmentCount = 2;
			PointF pt1 = GetLeftMiddle(link.Origin);
			PointF pt2 = GetTopMiddle(link.Destination);
			link.ControlPoints[0] = pt1;
			link.ControlPoints[1] = new PointF(pt2.X, pt1.Y);
			link.ControlPoints[2] = pt2;
			link.UpdateFromPoints();
		}

		private void BendLinkLeftDownRight(DiagramLink link)
		{
			link.CascadeOrientation = Orientation.Horizontal;
			link.Style = LinkStyle.Cascading;
			link.SegmentCount = 3;
			PointF pt1 = GetLeftMiddle(link.Origin);
			PointF pt2 = GetLeftMiddle(link.Destination);
			link.ControlPoints[0] = pt1;
			link.ControlPoints[1] = new PointF(pt1.X - 5, pt1.Y);
			link.ControlPoints[2] = new PointF(pt1.X - 5, pt2.Y);
			link.ControlPoints[3] = pt2;
			link.UpdateFromPoints();
		}

		private void BendLinkRightDown(DiagramLink link)
		{
			link.CascadeOrientation = Orientation.Horizontal;
			link.Style = LinkStyle.Cascading;
			link.SegmentCount = 2;
			PointF pt1 = GetRightMiddle(link.Origin);
			PointF pt2 = GetTopMiddle(link.Destination);
			link.ControlPoints[0] = pt1;
			link.ControlPoints[1] = new PointF(pt2.X, pt1.Y);
			link.ControlPoints[2] = pt2;
			link.UpdateFromPoints();
		}

		private void BendLinkRightDownLeft(DiagramLink link)
		{
			link.CascadeOrientation = Orientation.Horizontal;
			link.Style = LinkStyle.Cascading;
			link.SegmentCount = 3;
			PointF pt1 = GetRightMiddle(link.Origin);
			PointF pt2 = GetRightMiddle(link.Destination);
			link.ControlPoints[0] = pt1;
			link.ControlPoints[1] = new PointF(pt1.X + 5, pt1.Y);
			link.ControlPoints[2] = new PointF(pt1.X + 5, pt2.Y);
			link.ControlPoints[3] = pt2;
			link.UpdateFromPoints();
		}

		private void BendLinkDownLOR(DiagramLink link)
		{
			link.CascadeOrientation = Orientation.Vertical;
			link.Style = LinkStyle.Cascading;
			link.SegmentCount = 2;
			PointF pt1 = GetBottomMiddle(link.Origin);
			PointF pt2 = pt1.X > GetBottomMiddle(link.Destination).X ?
				GetRightMiddle(link.Destination) : GetLeftMiddle(link.Destination);
			link.ControlPoints[0] = pt1;
			link.ControlPoints[1] = new PointF(pt1.X, pt2.Y);
			link.ControlPoints[2] = pt2;
			link.UpdateFromPoints();
		}

		private void BendBackLink(DiagramLink link, float offset)
		{
			link.CascadeOrientation = Orientation.Horizontal;
			link.Style = LinkStyle.Cascading;
			link.SegmentCount = 3;
			PointF pt1 = GetRightMiddle(link.Origin);
			PointF pt2 = GetRightMiddle(link.Destination);
			float m = GetBottomMiddle(link.Destination).X;
			link.ControlPoints[0] = pt1;
			link.ControlPoints[1] = new PointF(m + offset + 5, pt1.Y);
			link.ControlPoints[2] = new PointF(m + offset + 5, pt2.Y);
			link.ControlPoints[3] = pt2;
			link.UpdateFromPoints();
		}

		private void SetLinkStraight(DiagramLink link)
		{
			link.Style = LinkStyle.Polyline;
			link.SegmentCount = 1;
			link.ControlPoints[0] = GetBottomMiddle(link.Origin);
			link.ControlPoints[1] = GetTopMiddle(link.Destination);
			link.UpdateFromPoints();
		}

		#endregion

		#region Drag and drop

		private void palette_MouseDown(object sender, MouseEventArgs e)
		{
			dragStartPoint = new Point(e.X, e.Y);
			dragItem = null;
		}

		private void palette_MouseMove(object sender, MouseEventArgs e)
		{
			if (palette.SelectedIndex == -1 || e.Button != MouseButtons.Left)
				return;

			if (dragItem == null)
				dragItem = palette.Items[palette.SelectedIndex];

			if (Math.Abs(dragStartPoint.X - e.X) > SystemInformation.DragSize.Width ||
				Math.Abs(dragStartPoint.Y - e.Y) > SystemInformation.DragSize.Height)
			{
				DoDragDrop(dragItem, DragDropEffects.Copy);
			}
		}

		private void view_DragOver(object sender, DragEventArgs e)
		{
			e.Effect = DragDropEffects.Copy;
			/*DiagramLink link = LinkFromDragEvent(e);
			if (link != null && (bool)link.Tag)
				e.Effect = DragDropEffects.Copy;
			else
				e.Effect = DragDropEffects.None;*/
		}

		private void view_DragDrop(object sender, DragEventArgs e)
		{
			var composite = diagram.UndoManager.StartComposite("Create");

			var point = view.ClientToDoc(view.PointToClient(new Point(e.X, e.Y)));
			var nodes = new DiagramNode[3];
			for (int i = 0; i < 3; i++)
				nodes[i] = diagram.Factory.CreateShapeNode(point.X + i * 50, point.Y, 30, 15);
			for (int i = 0; i < 3; i++)
				diagram.Factory.CreateDiagramLink(nodes[i], nodes[(i + 1) % 3]);

			new FlowchartLayout().Arrange(diagram);

			composite.Execute();

			/*e.Effect = DragDropEffects.None;

			DiagramLink link = LinkFromDragEvent(e);
			if (link != null && (bool)link.Tag)
			{
				e.Effect = DragDropEffects.Copy;

				string type = e.Data.GetData(typeof(string)).ToString();
				switch (type)
				{
				case "Activity":
					CreateActivity(link);
					break;
				case "IfElse":
					CreateIfElse(link);
					break;
				case "While":
					CreateWhile(link);
					break;
				case "Process":
					CreateProcess(link);
					break;
				}
			}*/
		}

		private DiagramLink LinkFromDragEvent(DragEventArgs e)
		{
			// DragEventArgs gives us the screen coordinates of a point
			// so convert them to the diagram logical coordinate system
			Point clientPoint = view.PointToClient(new Point(e.X, e.Y));
			PointF diagramPoint = view.ClientToDoc(clientPoint);

			return diagram.GetLinkAt(diagramPoint, 2);
		}

		#endregion


		#region Fields
		
		RectangleF defaultPosition = new RectangleF(0, 0, 35, 20);
		RectangleF smallSize = new RectangleF(0, 0, 5, 5);
		object dragItem = null;
		Point dragStartPoint = Point.Empty;

		#endregion

		#region Cosntants

		// layout traits keys
		private const string WorkflowItemId = "WF";
		private const string PairStart = "PS";
		private const string PairEnd = "PE";
		private const string Connector = "CN";

		// size constants
		private float LevelDistance = 26;

		#endregion

		private void btnUndo_Click(object sender, EventArgs e)
		{
			diagram.UndoManager.Undo();
		}

		private void btnRedo_Click(object sender, EventArgs e)
		{
			diagram.UndoManager.Redo();
		}
	}
}